#!/usr/bin/perl -w
#
# Name       : update-qd.pl
# Description: Update the qdinfo.dat file and qd.c source with any changes from
#              Stanfords project summary file, psummaryC.html
#              The script formally known as update-qd-files.pl
# Author     : Bas Couwenberg <sebastic@xs4all.nl>
# License    : GPL
#

use strict;
use warnings;
use File::Basename;
use Date::Calc;	
use LWP::UserAgent;
use Getopt::Mixed "nextOption";

# initialise configuration variables
my $self    = 'update-qd.pl';
my $version = '1.2.2';

# get todays date and convert the numerical month to text
my ($year, $m, $day, $hour, $min, $sec) = Date::Calc::Today_and_Now();
my $month   = Date::Calc::Month_to_Text($m);
my $date    = "$day $month $year";

$m    = '0'.$m    if(length($m)    == 1);
$day  = '0'.$day  if(length($day)  == 1);
$hour = '0'.$hour if(length($hour) == 1);
$min  = '0'.$min  if(length($min)  == 1);
$sec  = '0'.$sec  if(length($sec)  == 1);

my $logdate = "$year-$m-$day $hour:$min:$sec";

# epoch 2000 time in hexidecimal
my $hextime = &hextime();
   $hextime = sprintf("%08s", $hextime);

# lists to store the credits per project in
my %credit_of;
my %historic_credit_of;

my %number_of = (  pt_projects      => 0,
                   ph_projects      => 0,
 		   new_projects     => 0,
 		   updated_projects => 0
		);

my %cfg = ( 'psummary-url'    => 'http://fah-web.stanford.edu/psummaryC.html',
            'psummary-file'   => './psummaryC.html',
	    'qdinfo-file'     => './qdinfo.dat',
	    'qd-file'         => './qd.c',
	    'download'        => 1,
	    'dry-run'         => 0,
	    'force'           => 0,
	    'qdinfo-only'     => 0,

	    'compile'         => 0,
	    'compile-windows' => 1,
	    'compile-linux'   => 1,
	    'gcc-bin'         => '/usr/bin/gcc',
	    'mingw-bin'       => '/usr/bin/i586-mingw32msvc-gcc',
	    'windows-bin'     => './qd.exe',
	    'linux-bin'       => './qd',
		
	    'date'            => $date,
	    'hextime'         => $hextime,
	
	    'self'            => $self,
	    'version'         => $version,
	    'log'             => 0,
	    'log-file'        => './update-qd.log',
	    'log-data'        => '',
	    'verbose'         => 0
	  );

# get command line parameters to update the configuration
Getopt::Mixed::init("url=s
                     psummary=s
		     qdinfo=s
		     qd=s
		     verbose
		     help
		     compile
		     gcc=s
		     mingw=s
		     windows-bin=s
		     linux-bin=s
		     no-windows
		     no-linux
		     no-download
		     dry-run
		     force
		     qdinfo-only
		     log
		     log-file=s
		     u>url
		     p>psummary
		     i>qdinfo
		     q>qd
		     v>verbose
		     h>help
		     c>compile
		     g>gcc
		     m>mingw
		     w>windows-bin
		     l>linux-bin
		     nw>no-windows
		     nl>no-linux
		     nd>no-download
		     d>dry-run
		     f>force
		     io>qdinfo-only
		     L>log
		     Lf>log-file
		    ");

while (my ($option, $value) = nextOption()) {
        if($option eq 'help') {
		print_usage(%cfg);
		exit;
	}
	elsif($option eq 'url') {
		$cfg{'psummary-url'}    = $value;
	}
	elsif($option eq 'psummary') {
		$cfg{'psummary-file'}   = $value;
	}
	elsif($option eq 'qdinfo') {
		$cfg{'qdinfo-file'}     = $value;
	}
	elsif($option eq 'qd') {
		$cfg{'qd-file'}         = $value;
	}
	elsif($option eq 'verbose') {
		$cfg{'verbose'}         = 1;
	}
	elsif($option eq 'compile') {
		$cfg{'compile'}         = 1;
	}
	elsif($option eq 'gcc') {
		$cfg{'gcc-bin'}         = $value;
	}
	elsif($option eq 'mingw') {
		$cfg{'mingw-bin'}       = $value;
	}
	elsif($option eq 'windows-bin') {
		$cfg{'windows-bin'}     = $value;
	}
	elsif($option eq 'linux-bin') {
		$cfg{'linux-bin'}       = $value;
	}
	elsif($option eq 'no-windows') {
		$cfg{'compile-windows'} = 0;
	}
	elsif($option eq 'no-linux') {
		$cfg{'compile-linux'}   = 0;
	}
	elsif($option eq 'no-download') {
		$cfg{'download'}        = 0;
	}
	elsif($option eq 'dry-run') {
		$cfg{'dry-run'}         = 1;
	}
	elsif($option eq 'force') {
		$cfg{'force'}           = 1;
	}
	elsif($option eq 'qdinfo-only') {
		$cfg{'qdinfo-only'}     = 1;
	}
	elsif($option eq 'log') {
		$cfg{'log'}             = 1;
	}
	elsif($option eq 'log-file') {
		$cfg{'log-file'}        = $value;;
	}
}

# Step 0
# check the required permissions of the files
if(!&valid_permissions(%cfg)) {
	print "\nErrors encountered.\n";
	
	&abort();
}

# Step 1
# download the psummaryC.html from Stanford
if(!&download_psummary(%cfg)) {
	# abort further processing
	# we couldn't download the psummary file
	
	&abort();
}

# Step 2
# read the project/credit data from qdinfo.dat
%cfg = &parse_qdinfo(%cfg);

# Step 3
# update the project/credit data with any changes from psummaryC.html
%cfg = &update_projects(%cfg);

# if there are no new or updated projects
# we don't have to update qd.c and qdinfo.dat and can quite the program
# we do need to update if we are forced to
if($number_of{new_projects} == 0 && $number_of{updated_projects} == 0 && !$cfg{'force'}) {
	if($cfg{'verbose'}) {
		print "No data has changed, no need to update qdinfo.dat and qd.c.\n";

		&print_update_stats();
	}
	
	exit;
}
else {
	# if there were changes, log them if --log
	if($cfg{'log'}) {
		open(F, ">>", $cfg{'log-file'}) || die "Cannot open file: $cfg{'log-file'} ($!)";
		print F $cfg{'log-data'};
		close F;
	}
}

# Step 4
# generate updated qd.c and qdinfo.dat files
# skip qd.c and its possible compile if qdinfo-only is requested

my $qdinfo_data = &generate_qdinfo(%cfg);
my $qd_data     = &generate_qd(%cfg) if($cfg{'qdinfo-only'} != 1);

# Step 5
# Save the generated files

if($cfg{'dry-run'}) {
	# if we are to leave the orginal files alone 
	# overwrite the paths to qdinfo.dat and qd.c,
	# and the paths to the qd.exe and qd binaries

	$cfg{'qdinfo-file'} = './new-qdinfo.dat';
	$cfg{'qd-file'}     = './new-qd.c';

	$cfg{'windows-bin'} = './new-qd.exe';
	$cfg{'linux-bin'}   = './new-qd';
}

print "\nSaving updated qdinfo.dat to: $cfg{'qdinfo-file'}\n";

open(F, ">", $cfg{'qdinfo-file'}) || die "Cannot open file: $cfg{'qdinfo-file'} ($!)";
print F $qdinfo_data;
close F;

if($cfg{'qdinfo-only'} != 1) {
	print "\nSaving updated qd.c to: $cfg{'qd-file'}\n";

	open(F, ">", $cfg{'qd-file'}) || die "Cannot open file: $cfg{'qd-file'} ($!)";
	print F $qd_data;
	close F;

	# Step 6 (Optional)
	# Compile qd binary from qd
	if($cfg{'compile'} == 1) {
		&compile_qd(%cfg);
	}
}

&print_update_stats();

################################[ Subroutines ]################################

sub abort {
	print "\nProgram aborted!\n";

	exit 1;
}

sub compile_qd {
	my %cfg = @_;

	if($cfg{'compile-linux'} == 1) {
		print "\n";
		print "Compiling qd binary for Linux ... \n";
		print "Command: $cfg{'gcc-bin'} -Wall -DSYSTYPE=0 -s -O2 -o $cfg{'linux-bin'} $cfg{'qd-file'}\n";
		
		system($cfg{'gcc-bin'}, '-Wall', '-DSYSTYPE=0', '-s', '-O2', '-o', $cfg{'linux-bin'}, $cfg{'qd-file'});
	}
	if($cfg{'compile-windows'} == 1) {
		print "\n";
		print "Compiling qd binary for Windows ... \n";
		print "Command: $cfg{'mingw-bin'} -Wall -DSYSTYPE=1 -s -O2 -o $cfg{'windows-bin'} $cfg{'qd-file'}\n";
			
		system($cfg{'mingw-bin'}, '-Wall', '-DSYSTYPE=1', '-s', '-O2', '-o', $cfg{'windows-bin'}, $cfg{'qd-file'});
	}
}

sub print_update_stats {
	print "\n";
	print "Total number of projects in qdinfo.dat : ";
	print $number_of{pt_projects};
	print " (+ ";
	print $number_of{new_projects};
	print ")\n";
	print "Total number of point history entries  : ";
	print $number_of{ph_projects};
	print " (+ ";
	print $number_of{updated_projects};
	print ")\n";
}

sub generate_qd {
	my %cfg = @_;

	my $qd_head = '';
	my $qd_pt   = '';
	my $qd_tail = '';

	my $parse   = 0;
	open(F, $cfg{'qd-file'}) || die "Cannot open file: $cfg{'qd-file'} ($!)";
	while(<F>) {
		if($parse == 0 && $qd_pt eq '' && $_ !~ /\/\* The following text is automatically generated from \<qdinfo\.dat\> \*\//) {
			$qd_head .= $_;
		}
		elsif(($parse == 1 && $_ !~ /};/) || /\/\* The following text is automatically generated from \<qdinfo\.dat\> \*\//) {
			$parse  = 1;
			
			$qd_pt .= $_;
		}
		elsif($parse == 1 && $_ =~ /};/) {
			$parse  = 0;
			$qd_pt .= $_;
		}
		else {
			$qd_tail .= $_;
		}
	}
	close F;

	$qd_pt  = "/* The following text is automatically generated from <qdinfo.dat> */\n\n";
	$qd_pt .= "#define DATE \"$cfg{'date'}\"\n";
	$qd_pt .= "#define FREV \"$cfg{'functional-revision'}\"\n";
	$qd_pt .= "#define PGEN 0x$cfg{'hextime'}\n";
	$qd_pt .= "struct ptent ptbl[] =\t\t/* Built-in table, from \"pt\" and \"ph\" entries */\n";
	   
	my $i = 0;
	my $p = 0;
	foreach my $project (sort { $a <=> $b } keys %credit_of) {
		# example line:
		#         {  127,  200 },{  128,  200 },{  129,  200 },{  131,  140 },{  132,  280 },
		
		if($i == 0) {
			# first line in the structure, needs opening curly brace
			$qd_pt .= " {\t";
		}
	
		$qd_pt .= "{ ";
		$qd_pt .= sprintf("%4s", $project);
		$qd_pt .= ",";
		$qd_pt .= sprintf("%6s", $credit_of{$project});
		$qd_pt .= " },";
		
		if($p == 4) {
			$qd_pt .= "\n\t";
			$p = -1;
		}
		
		$i++;
		$p++;
	}
	$qd_pt .= "\n\t" if($qd_pt !~ /\n\t$/);
	
	$p = 0;
	foreach my $project (sort { $a <=> $b } keys %historic_credit_of) {
		# example line:
		#         {  171,  200, 0x0A7A948C },{  180,  560, 0x0818538C },{  182,  560, 0x0818538C },
		
		foreach my $project_history (@{$historic_credit_of{$project}}) {
			my ($credits, $hextime) = split /:/, $project_history;
			
			$qd_pt .= "{ ";
			$qd_pt .= sprintf("%4s", $project);
			$qd_pt .= ",";
			$qd_pt .= sprintf("%6s", $credits);
			$qd_pt .= ", 0x$hextime },";
			
			if($p == 2) {
				$qd_pt .= "\n\t";
				$p = -1;
			}
			
			$p++;
		}
	}
	
	$qd_pt .= "\n\t" if($qd_pt !~ /\n\t$/);
	$qd_pt .= "{ 0 }\n";
	$qd_pt .= " };\n";

	my $qd_data = $qd_head.$qd_pt.$qd_tail;

	return $qd_data;
}

sub generate_qdinfo {
	my %cfg = @_;	

	# define the components of the new qdinfo.dat
	# the qdinfo.dat file has 3 sections:
	# - the header        with the functional revision, the date it was generated
	#                     both as human readable date as well as epoch 2000
	#                     timestamp, and by whom the file was generated
	# - the point table   with the current credits per project
	# - the point history with the credit revisions per project

	my $qdinfo_head   = "# <qdinfo.dat>\n\n";
	   $qdinfo_head  .= "da $cfg{'date'} ($cfg{'self'})\n\n";
	   $qdinfo_head  .= "# Latest functional revision number of \"qd\"\n\n";
	   $qdinfo_head  .= "fr $cfg{'functional-revision'}\n\n";
	   $qdinfo_head  .= "# Point table generation time, epoch 2000\n\n";
	   $qdinfo_head  .= "pg $cfg{'hextime'}\n";
	my $point_table   = "\n# Point table entries (points per WU (project, 100 * points))\n\n";
	my $point_history = "\n# Point history entries (project, 100 * old points, change date epoch 2000)\n\n";

	# recreate point table with updated project:value combinations
	# each row starts with pt and should have 5 project:value combinations
	my $i  = 0;
	foreach my $project (sort { $a <=> $b } keys %credit_of) {
		# example pt line:
		# pt  101  7400   111   500   112   500   113   500   114   500
	
		if($i == 0) {
			$point_table .= "pt ";
		}
		if($i <= 4) {
			$point_table .= sprintf("%4s", $project);
			$point_table .= " ";
			$point_table .= sprintf("%6s", $credit_of{$project});
			$point_table .= "  " if($i < 4);
		}
		if($i == 4) {
			$point_table .= "\n";
			$i = -1;
		}
		$i++;
	}
	# last pt row possible does not contain 5 projects
	if($point_table =~ s/  $/\n/) {}
	
	# recreate point history table with updated credit revisions
	# each row starts with ph and should have a max of 3 credit revisions per line
	# if a project has more than 3 credit revisions addional ph lines need to be
	# written to list them all
	
	foreach my $project (sort { $a <=> $b } keys %historic_credit_of) {
		# example ph line:
		# ph 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		
		$point_history .= "ph ";
		my $i = 0;
		foreach my $project_history (@{$historic_credit_of{$project}}) {
			my ($credits, $hextime) = split /:/, $project_history;
			
			# if this is not the first revision written on the line
			# add 2 spaces as a separator
			if($i > 0) {
				$point_history .= "  ";
			}
	
			$point_history .= sprintf("%4s", $project);
			$point_history .= " ";
			$point_history .= sprintf("%6s", $credits);
			$point_history .= " ";
			$point_history .= sprintf("%08s", $hextime);
			
			if($i == 2) {
				# wrote 3 revistion on one line
				# maximum reached
				
				$point_history .= "\n";
				
				if($#{$historic_credit_of{$project}} > 2) {
					# if this project has more than 3 revisions
					# write a new point history line
					
					$point_history .= "ph ";
				}
				
				$i = -1; # reset the field counter
			}
	
			$i++;
		}
		# add a newline if we didn't write a full line with 3 revisions
		$point_history .= "\n" unless($point_history =~ /\n$/);
	}

	# compose new qdinfo.dat file
	my $qdinfo_data  = $qdinfo_head.$point_table.$point_history;

	return $qdinfo_data;
}

sub update_projects {
	my %cfg = @_;

	print "\nUpdating projects ...\n\n" if($cfg{'verbose'});

	open (F, "<", $cfg{'psummary-file'}) || die "Cannot open file: $cfg{'psummary-file'} ($!)";
	while (<F>) {                           # read from file
		# Columns per tablerow:
		#
		#  0 Project Number
		#  1 Server IP
		#  2 Work Unit Name
		#  3 Number of Atoms
		#  4 Preferred (days)
		#  5 Final Deadline (days)
		#  6 Credit
		#  7 Frames
		#  8 Code
		#  9 Description
		# 10 Contact
		#	
		# example line to parse:
		#
		#<TR bgcolor="#dfdfdf"> <TD>1311</TD><TD> 171.65.103.154 </TD>
		#<TD>p1311_1fnt_a10_91ps_CMnone</TD><TD>90068</TD><TD>   25.00</TD>
		#<TD>   37.00</TD><TD>  254.00</TD><TD>100</TD><TD>GROMACS</TD>
		#<TD><a href=http://fah-web.stanford.edu/cgi-bin/allprojects#1311>Description</a></TD>
		#<TD><font size=-1>sungjoo</font></TD></TR>
		
		next unless /cgi-bin/;          # only parse project lines
		chomp;                          # remove start/end junk
		
		s/<tr.*?>\s*<td>\s*//ig;        # strip <TR bgcolor="#dfdfdf"> <TD>
		s/\s*<\/td>\s*?<td>\s*/\;/ig;   # replace all </td><td> tags by ;
		s/\s*<\/td>\s*<\/tr>\s*//ig;    # strip </TD></TR>
	
		# example line;
		#243;171.64.122.127;p243_vil8MUreGS;6372;16.00;23.00;78.00;100;GROMACS
		#;<a href=http://fah-web.stanford.edu/cgi-bin/allprojects#243>Description</a>
		#;<font size=-1>rbrandman</font>
		
		my ($project, $credits) = (split(";", $_))[0,6]; # get data split by ;
		$credits  = $credits + 0;       # store\
		$credits *= 100;                # multiply by 100,
		                                # value is saved in qdinfo like that
		
		if($credit_of{$project}) {
			# the project exists in qdinfo.dat
			
			# update the credit value 
			# if it is different from the credits in qdinfo.dat
			if($credit_of{$project} != $credits) {
				my $output  = "Updated project : ";
				   $output .= sprintf("%4s", $project);
				   $output .=  "  new credits*100: ";
				   $output .= sprintf("%6s", $credits);
				   $output .=  "  old credits*100: ";
				   $output .= sprintf("%6s", $credit_of{$project});
				   $output .= "\n";
					
				#print  "Updated project : ";
				#printf("%4s", $project);
				#print  "  new credits*100: ";
				#printf("%6s", $credits);
				#print  "  old credits*100: ";
				#printf("%6s", $credit_of{$project});
				#print "\n";

				print $output;
				if($cfg{'log'}) {
					$cfg{'log-data'} .= "[$logdate] ".$output;
				}
				
				# add the old credit value at the beginning 
				# of the revision list of this project
				unshift @{$historic_credit_of{$project}}, "$credit_of{$project}:$cfg{'hextime'}";
				
				# update the amount of credits for this project in the list
				$credit_of{$project} = $credits;
				
				# increase the counter of updated projects
				$number_of{updated_projects} += 1;
			}
		}
		else {
			# the project does not exist in qdinfo.dat yet
			
			my $output  = "    New project : ";
			   $output .= sprintf("%4s", $project);
			   $output .= "  new credits*100: ";
			   $output .= sprintf("%6s", $credits);
			   $output .= "\n";
			
			#print  "    New project : ";
			#printf("%4s", $project);
			#print  "  new credits*100: ";
			#printf("%6s", $credits);
			#print  "\n";
			
			print $output;
			if($cfg{'log'}) {
				$cfg{'log-data'} .= "[$logdate] ".$output;
			}
			
			# add this project to the list
			$credit_of{$project} = $credits;
			
			# increase the counter of new projects
			$number_of{new_projects} += 1;
		}
	}	
	close F;

	return %cfg;
}

sub parse_qdinfo {
	my %cfg = @_;
	
	# the functional revision is very important
	# if we're unable to get it from qdinfo.dat we should abort
	my $seen_fr = 0;
	
	open(F, $cfg{'qdinfo-file'}) || die "Cannot open file: $cfg{'qdinfo-file'} ($!)";
	while(<F>) {
		# only parse functional revision, point table and point history
	        next unless /^(fr|pt|ph)/; 
		
		# example funtional revision line:
		# fr 032
		if(/^fr (\d+)/) {
			$cfg{'functional-revision'} = $1;
		
			$seen_fr = 1;
		}
		# example point table lines:
		# pt 2051 16700  2052 16600  2053 13700  2054  6900  2055 12900
		# pt 2056 12800  2057 10600  2059  3800  2060  5700
		elsif(/^pt/) {
			s/^pt\s+//; # strip pt prefix
	
			# The point table lines contain dubble space separated pairs of
			# project and credit data.
			# The first thing on each line are the pt characters to
			# indicate that this line contains point table data.
			# The rest of the line is populated with the actual data, first
			# the project number followed by a space and then the amount of
			# credits for that project multiplied by 100
			#
			# To parse these lines, I remove the pt prefix, and create a
			# list of values by splitting the data on the line on one or
			# more spaces. To keep track of which value in the list is the
			# project number and which is the amount of credits, the
			# variable $i flags the data.
	
			my $i = 0;
			my $project = '';
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the projectnumber
					
					# save the project number
					$project = $_;
	
					# increase the project counter
					$number_of{pt_projects} += 1;
	
					# next value in the list are the credits
					$i = 1;
				}
				else {
					# second value in the pair
					# are the credits of the project
					
					# add the credits to to the list
					# which we will use later to generate
					# the new qdinfo.dat file
	
					$credit_of{$project} = $_;
					
					# next value in the list is the project
					$i = 0;
				}
			}
		}
		# example point history lines:
		# ph 2018 15400 0A584D8C
		# ph 2019 20800 0A77D56C  2019 20100 0A72734C  2019 11600 0A71E6AC
		elsif(/^ph/) {
			# strip carriage return
			if(s/\r//) {}
		
			s/^ph\s+//; # strip ph prefix
	
			# The point history lines contain groups of credit revision 
			# history separted by double spaces. 
			# The first thing on each line are the ph characters to
			# indicate that this line contains point history data.
			# The rest of the line is populated with the actual data, first 
			# the project number followed by a space, then the amount of 
			# credits multiplied by 100, a space, and the epoch 2000 time
			# which indicated when the amount of credits was changed to 
			# that value.
			# Every line contains one or more credit revisions of only one
			# project. Because of the maximum of 80 character per line, 
			# there are at most 3 revision on one line.
			# If a project has more than 3 credit revision, additional 
			# ph lines are created to list all the revisions.
			# So it is very well posible to see a specific project which 
			# has several ph lines.
			# The revisions are ordered from latest up to the oldest. So 
			# the most recent revision is at the beginning of the line.
			#
			# To parse these lines, I remove the ph prefix, and create a
			# list of values by splitting the data on the line on one or 
			# more spaces. To keep track of which value in the list is the 
			# project numer, which is the amount of credits and which is 
			# the revision epoch 2000 time, the variable $i flags the data.
			
			my ($project, $credits, $hextime) = ('','','');
			my $i = 0;
			foreach(split /\s+/) {
				if($i == 0) {
					# the first value in the pair
					# is the project number
					
					# save the project number
					$project = $_;
					
					# increase the project counter
					$number_of{ph_projects} += 1;
				}
				elsif($i == 1) {
					# second value in the pair
					# is the credits*100 of the project
					
					# save the amount of credits
					$credits = $_;				
				}
				elsif($i == 2) {
					# third value in the pair
					# is the date of the change in epoch 2000
					# from this machines perspective
					
					# save the epoch 2000 time
					$hextime = $_;
					
					# add the revision data to the list 
					# for this project
					push @{$historic_credit_of{$project}}, "$credits:$hextime";
	
					# reset field counter/flag
					$i = -1; 
				}
				$i++;
			}
		}
	}	
	close F;
	
	if($seen_fr) {
		return %cfg;
	}
	else {
		print "Unable to find the functional revision in the file!\n";
		print "Please verify the \"fr\" entry in qdinfo.dat!\n";

		&abort();
	}
}

sub download_psummary {
	my %cfg = @_;
	
	# skip the download if requested
	if(!$cfg{'download'}) {
		if($cfg{'verbose'}) {
			print "Skipping download of URL: $cfg{'psummary-url'}\n";
		}
		
		# check if there is actually a file on disk
		# abort if missing/unreadable
		if(-r $cfg{'psummary-file'}) {
			print "Using file already on disk: $cfg{'psummary-file'}\n" if($cfg{'verbose'});
		}
		else {
			print "Error: Cannot read psummary file! ($cfg{'psummary-file'})\n";
			print "Cannot continue without psummary file!\n";
			
			&abort();
		}
		
		return 1;
	}
		
	print "Downloading URL: $cfg{'psummary-url'} ... " if($cfg{'verbose'});

	# create useragent object which will handle the HTTP request
	my $ua  = LWP::UserAgent->new();
	   $ua->agent("$cfg{self}/$cfg{version}");
	
	# create and send the HTTP request
	my $req = HTTP::Request->new(GET => $cfg{'psummary-url'});
	my $res = $ua->request($req);
	if($res->is_success) {
		print "Success\n" if($cfg{'verbose'});
		
		print "Saving file to disc: $cfg{'psummary-file'}\n" if($cfg{'verbose'});
		open(F, ">", $cfg{'psummary-file'}) || die "Cannot open file: $cfg{'psummary-file'} ($!)";
		print F $res->content;
		close F;

		return 1;
	}
	else {
		print "Failed!\n" if($cfg{'verbose'});
		print "Error: Failed to retrieve URL! ($cfg{'psummary-url'})\n";
		print "Server returned: ".$res->code." ".$res->message."\n";
		
		return undef;
	}
}

sub valid_permissions {
	my %cfg = @_;

	print "Checking file permissions\n" if($cfg{'verbose'});

	# psummaryC.html
	# only check the file if it exist
	# check the parent directory if it has yet to be created
	#
	# needs read/write permissions to the file
	# or write permissions to the directory in which it is to be created
	
	print " psummaryC.html : " if($cfg{'verbose'});
	if(-e $cfg{'psummary-file'}) {
		if(!-w $cfg{'psummary-file'} || !-r $cfg{'psummary-file'}) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'psummary-file'}\n";
			
			if(!-r $cfg{'psummary-file'}) {
				print "   Cannot read file!\n";
			}
			if(!-w $cfg{'psummary-file'}) {
				print "   Cannot write file!\n";
			}

			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}
	}
	else {
		my $dir = dirname($cfg{'psummary-file'});
		if(!-w $dir) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'psummary-file'}\n";
			print "   Cannot write directory!\n";
			
			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}
	}

	# qdinfo.dat
	# needs read/write permissions
	print " qdinfo.dat     : " if($cfg{'verbose'});
	if(!-r $cfg{'qdinfo-file'} || !-w $cfg{'qdinfo-file'}) {
		print "Error!\n" if($cfg{'verbose'});
		print "  file: $cfg{'qdinfo-file'}\n";
		if(!-r $cfg{'qdinfo-file'}) {
			print "   Cannot read file!\n";
		}
		if(!-w $cfg{'qdinfo-file'}) {
			print "   Cannot write file!\n";
		}
		
		return undef;
	}
	else {
		print "OK\n" if($cfg{'verbose'});
	}
	
	# qd.c
	# needs read/write permissions
	# check only if qdinfo-only is not requested
	if($cfg{'qdinfo-only'} != 1) {
		print " qd.c           : " if($cfg{'verbose'});
		if(!-r $cfg{'qd-file'} || !-w $cfg{'qd-file'}) {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: $cfg{'qd-file'}\n";
			if(!-r $cfg{'qd-file'}) {
				print "   Cannot read file!\n";
			}
			if(!-w $cfg{'qd-file'}) {
				print "   Cannot write file!\n";
			}
		
			return undef;
		}
		else {
			print "OK\n" if($cfg{'verbose'});
		}

		# if we need to compile, check those files and apps too
		if($cfg{'compile'} == 1) {
			if($cfg{'compile-linux'} == 1) {
				# check the linux specific parameters
				
				print " gcc            : " if($cfg{'verbose'});
				if(!-x $cfg{'gcc-bin'}) {
					print "Error!\n" if($cfg{'verbose'});
					
					print "  file: $cfg{'gcc-bin'}\n";
					print "   Cannot execute file!\n";
				}
				else {
					print "OK\n" if($cfg{'verbose'});
				}
	
				print " qd             : " if($cfg{'verbose'});
				if(-e $cfg{'linux-bin'}) {
					if(!-w $cfg{'linux-bin'}) {
						print "Error!\n" if($cfg{'verbose'});
						
						print "  file: $cfg{'linux-bin'}\n";
						print "   Cannot write file!\n";
	
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
				else {
					my $dir = dirname($cfg{'linux-bin'});
					if(!-w $dir) {
						print "Error!\n" if($cfg{'verbose'});
						print "  file: $cfg{'linux-bin'}\n";
						print "   Cannot write directory!\n";
				
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
			}
			if($cfg{'compile-windows'}) {
				# check the windows specific parameters
				
				print " mingw          : " if($cfg{'verbose'});
				if(!-x $cfg{'mingw-bin'}) {
					print "Error!\n" if($cfg{'verbose'});
					
					print "  file: $cfg{'mingw-bin'}\n";
					print "   Cannot execute file!\n";
				}
				else {
					print "OK\n" if($cfg{'verbose'});
				}
	
				print " qd.exe         : " if($cfg{'verbose'});
				if(-e $cfg{'windows-bin'}) {
					if(!-w $cfg{'windows-bin'}) {
						print "Error!\n" if($cfg{'verbose'});
							
						print "  file: $cfg{'windows-bin'}\n";
						print "   Cannot write file!\n";
	
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
				else {
					my $dir = dirname($cfg{'windows-bin'});
					if(!-w $dir) {
						print "Error!\n" if($cfg{'verbose'});
						print "  file: $cfg{'windows-bin'}\n";
						print "   Cannot write directory!\n";
				
						return undef;
					}
					else {
						print "OK\n" if($cfg{'verbose'});
					}
				}
			}
		}
	}

	# update-qd log file
	# needs read/write permissions
	if($cfg{'log'}) {
		print " update-qd log  : " if($cfg{'verbose'});
		if($cfg{'log-file'}) {
			if(-e $cfg{'log-file'}) {
				if(-r $cfg{'log-file'} && -w $cfg{'log-file'}) {
					print "OK\n" if($cfg{'verbose'});
				}
				else {
					print "Error!\n" if($cfg{'verbose'});
					print "  file: $cfg{'log-file'}\n";
					if(!-r $cfg{'log-file'}) {
						print "   Cannot read file!\n";
					}
					if(!-w $cfg{'log-file'}) {
						print "   Cannot write file!\n";
					}
					
					return undef;
				}
			}
			else {
				if(-w dirname($cfg{'log-file'})) {
					print "OK\n" if($cfg{'verbose'});
				}
				else {
					print "Error!\n" if($cfg{'verbose'});
					print "  file: $cfg{'log-file'}\n";
					print "   Cannot write directory!\n";
				}
			}
		}
		else {
			print "Error!\n" if($cfg{'verbose'});
			print "  file: none\n";
			print "   No log-file specified!\n";

			return undef;
		}
	}
	
	print "\n" if($cfg{'verbose'});

	return 1;
}

sub decode_hextime {
	# convert a hexadecimal epoch 2000 value to human readable date string
	
	# Some notes on epoch 2000 vs epoch 1970
	#
	# epoch 2000/Cosm
	# starts: January 1, 2000, 00:00:00 UTC
	#
	# epoch 1970/UNIX
	# starts: January 1, 1970, 00:00:00 UTC
	#
	# difference between the two dates: 946684800 seconds
	#
	# example epoch 2000 strings:
	# 0A8DA04A
	# 0817020C
	#
	# $date_string = scalar localtime (hex("0A8DA04A") + 946684800);
	#
	# add the seconds between epoch 2000 and epoch 1970 to the decimal
	# value of the hex date to get the total amount of seconds since the
	# UNIX epoch convert that to human readable date through scalar
	# localtime
	#
	
	my $hextime = $_[0];

	# scalar localtime hex($_), uses the timezone of this machine
	# not Dick Howells who originally generated the qdinfo.dat files
	
	return scalar localtime (hex($hextime) + 946684800);
}

sub hextime {
	# convert current date to hex
	
	my $time  = time;                 # seconds since epoch 1970
	   $time -= 946684800;            # seconds since epoch 2000
	   $time  = sprintf("%X", $time); # seconds to hex in uppercase
	   
	return $time;
}

sub print_usage {
	my %cfg = @_;

	$cfg{'verbose'}         = $cfg{'verbose'}         == 1 ? 'true' : 'false';
	$cfg{'download'}        = $cfg{'download'}        == 1 ? 'true' : 'false';
	$cfg{'dry-run'}         = $cfg{'dry-run'}         == 1 ? 'true' : 'false';
	$cfg{'force'}           = $cfg{'force'}           == 1 ? 'true' : 'false';
	$cfg{'qdinfo-only'}     = $cfg{'qdinfo-only'}     == 1 ? 'true' : 'false';
	$cfg{'compile'}         = $cfg{'compile'}         == 1 ? 'true' : 'false';
	$cfg{'compile-linux'}   = $cfg{'compile-linux'}   == 0 ? 'true' : 'false';
	$cfg{'compile-windows'} = $cfg{'compile-windows'} == 0 ? 'true' : 'false';
	$cfg{'log'}             = $cfg{'log'}             == 1 ? 'true' : 'false';
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	
	print "-u, --url <url>           URL to psummaryC.html\n";
	print "                           default: $cfg{'psummary-url'}\n";
	print "-p, --psummary <file>     Path to save psummaryC.html to\n";
	print "                           default: $cfg{'psummary-file'}\n";
	print "-q, --qd <file>           Path to qd.c file\n";
	print "                           default: $cfg{'qd-file'}\n";
	print "-i, --qdinfo <file>       Path to qdinfo.dat file\n";
	print "                           default: $cfg{'qdinfo-file'}\n";
	print "--nd, --no-download       Don't download the psummaryC.html file\n";
	print "                           default: $cfg{'download'}\n";
	print "--io, --qdinfo-only       Only update qdinfo.dat, skip qd.c\n";
	print "                           default: $cfg{'qdinfo-only'}\n";
	print "-f, --force               Force regeneration of point table\n";
	print "                           default: $cfg{'force'}\n";
	print "-d, --dry-run             Don't overwrite the qd.c and qdinfo.dat\n";
	print "                          Will save files to ./new-qd.c and ./new-qdinfo.dat\n";
	print "                          Will compile to ./new-qd.exe and ./new-qd\n";
	print "                           default: $cfg{'dry-run'}\n";
	print "-L, --log                 Log new and/or updated projects.\n";
	print "                           default: $cfg{'log'}\n";
	print "--Lf, --log-file <file>   Path to log file.\n";
	print "                           default: $cfg{'log-file'}\n";
	
	print "\n";	
	print "-v, --verbose             Enable verbose output\n";
	print "                           default: $cfg{'verbose'}\n";
	print "-h, --help                Print usage information\n";
	
	print "\nCompile Options:\n";
	print "-c, --compile             Enable compilation of updated qd.c\n";
	print "                           default: $cfg{'compile'}\n";
	print "-g, --gcc <file>          Path to gcc executable\n";
	print "                           default: $cfg{'gcc-bin'}\n";
	print "-m, --mingw <file>        Path to mingw executable\n";
	print "                           default: $cfg{'mingw-bin'}\n";
	print "-l, --linux-bin <file>    Path to the Linux qd binary\n";
	print "                           default: $cfg{'linux-bin'}\n";
	print "-w, --windows-bin <file>  Path to the Windows qd binary\n";
	print "                           default: $cfg{'windows-bin'}\n";
	print "--nl, --no-linux          Don't compile the Linux binary\n";
	print "                           default: $cfg{'compile-linux'}\n";
	print "--nw, --no-windows        Don't compile the Windows binary\n";
	print "                           default: $cfg{'compile-windows'}\n";
}
